@@ -1,5 +1,3 @@
-## $Id: npregiv.R,v 1.8 2011/05/30 19:17:35 jracine Exp jracine $
-
 ## This functions accepts the following arguments:
 
 ## y: univariate outcome
@@ -22,8 +20,13 @@
 
 ## This function returns a list with the following elements:
 
-## phihat: the IV estimator of phi(y)
+## phi: the IV estimator of phi(y)
 ## alpha:  the Tikhonov regularization parameter
+## phi.mat: the matrix with colums phi_1, phi_2 etc. over all iterations
+## num.iterations:  the number of Landweber-Fridman iterations
+## norm.stop: the vector of values of the objective function used for stopping
+## norm.value: the norm not multiplied by the number of iterations
+## convergence: a character string indicating whether/why iteration terminated
 
 npregiv <- function(y,
                     z,
@@ -33,14 +36,26 @@ npregiv <- function(y,
                     weval=NULL,
                     xeval=NULL,
                     p=1,
+                    nmulti=1,
+                    random.seed=42,
+                    optim.maxattempts = 10,
+                    optim.method=c("Nelder-Mead", "BFGS", "CG"),
+                    optim.reltol=sqrt(.Machine$double.eps),
+                    optim.abstol=.Machine$double.eps,
+                    optim.maxit=500,
                     alpha=NULL,
                     alpha.min=1.0e-10,
                     alpha.max=1.0e-01,
                     alpha.tol=.Machine$double.eps^0.25,
-                    iterate.max=100,
-                    iterate.tol=1.0e-05,
+                    iterate.max=1000,
+                    iterate.diff.tol=1.0e-08,
                     constant=0.5,
                     method=c("Landweber-Fridman","Tikhonov"),
+                    penalize.iteration=TRUE,
+                    smooth.residuals=TRUE,
+                    start.from=c("Eyz","EEywz"),
+                    starting.values=NULL,
+                    stop.on.increase=TRUE,
                     ...) {
 
   ## This function was constructed initially by Samuele Centorrino
@@ -96,7 +111,7 @@ npregiv <- function(y,
   ## phi:   the vector of estimated values for the unknown function at the evaluation points
   
   tikh <- function(alpha,CZ,CY,Cr.r){
-    return(solve(alpha*diag(length(Cr.r)) + CY%*%CZ) %*% Cr.r)
+    return(chol2inv(chol(alpha*diag(length(Cr.r)) + CY%*%CZ) %*% Cr.r))
   }
   
   ## This function applies the iterated Tikhonov approach which
@@ -128,7 +143,7 @@ npregiv <- function(y,
   ## which is a function of alpha (see (3.10) of Feve & Florens (2010)
   
   ittik <- function(alpha,CZ,CY,Cr.r,r) {
-    invmat <- solve(alpha*diag(length(Cr.r)) + CY%*%CZ)
+    invmat <- chol2inv(chol(alpha*diag(length(Cr.r)) + CY%*%CZ))
     phi <- invmat %*% Cr.r + alpha * invmat %*% invmat %*% Cr.r        
     return(sum((CZ%*%phi - r)^2)/alpha)    
   }
@@ -152,8 +167,9 @@ npregiv <- function(y,
                       bws=NULL,
                       p=0,
                       shrink=TRUE,
-                      warning.immediate=TRUE) {
-  
+                      warning.immediate=TRUE,
+                      ...) {
+
     ## Basic error checking...
   
     if(is.null(mydata.train)) stop("You must provide training data")
@@ -253,14 +269,15 @@ npregiv <- function(y,
           }
         }
       }
-  
+
       if(p == 0) {
   
         Wmat.sum <- npksum(exdat=X.eval[j,],
                            txdat=X.train,
                            bws=bws,
                            ukertype="liracine",
-                           okertype="liracine")$ksum
+                           okertype="liracine",
+                           ...)$ksum
   
       } else {
   
@@ -270,7 +287,8 @@ npregiv <- function(y,
                            weights=W,
                            bws=bws,
                            ukertype="liracine",
-                           okertype="liracine")$ksum[,,1]
+                           okertype="liracine",
+                           ...)$ksum[,,1]
   
       }
   
@@ -282,7 +300,8 @@ npregiv <- function(y,
                   exdat=X.train, 
                   bws=bws,
                   ukertype="liracine",
-                  okertype="liracine")$ksum
+                  okertype="liracine",
+                  ...)$ksum
   
       ## p == 0
   
@@ -292,7 +311,7 @@ npregiv <- function(y,
   
       ## No singularity problems...
   
-      if(tryCatch(Wmat.sum.inv <- as.matrix(solve(Wmat.sum)),
+      if(tryCatch(Wmat.sum.inv <- as.matrix(chol2inv(chol(Wmat.sum))),
                   error = function(e){
                     return(matrix(FALSE,nc,nc))
                   })[1,1]!=FALSE) {
@@ -317,14 +336,14 @@ npregiv <- function(y,
           epsilon <- 1/n.train
           ridge <- 0
           
-          while(tryCatch(as.matrix(solve(Wmat.sum+diag(rep(ridge,nc)))),
+          while(tryCatch(as.matrix(chol2inv(chol((Wmat.sum+diag(rep(ridge,nc)))))),
                          error = function(e){
                            return(matrix(FALSE,nc,nc))
                          })[1,1]==FALSE) {
             ridge <- ridge + epsilon
           }
   
-          Wmat.sum.inv <- as.matrix(solve(Wmat.sum+diag(rep(ridge,nc))))
+          Wmat.sum.inv <- as.matrix(chol2inv(chol(Wmat.sum+diag(rep(ridge,nc)))))
   
           ## Add for debugging...
   
@@ -417,12 +436,6 @@ npregiv <- function(y,
   
   ## Functions for generalized local polynomial regression
   
-  ## No Zero Denominator, used in C code for kernel estimation...
-  
-  NZD <- function(a) {
-    sapply(1:NROW(a), function(i) {if(a[i] < 0) min(-.Machine$double.xmin,a[i]) else max(.Machine$double.xmin,a[i])})
-  }
-  
   mypoly <- function(X,degree) {
   
     if(missing(X)) stop(" X required")
@@ -456,13 +469,13 @@ npregiv <- function(y,
       xdat.numeric <- as.data.frame(xdat[,xdat.col.numeric])
     }
   
-    if(length(degree) != ncol(xdat.numeric)) stop(" degree vector and number of numeric predictors incompatible")
+    if(length(degree) != k) stop(" degree vector and number of numeric predictors incompatible")
   
     if(all(degree == 0) | k == 0) {
   
       ## Local constant OR no continuous variables
-      
-      return(matrix(1,nrow=nrow(xdat.numeric),ncol=1))
+
+      return(matrix(1,nrow=nrow(xdat),ncol=1))
       
     } else {
   
@@ -482,7 +495,7 @@ npregiv <- function(y,
           }
         }
       }
-      res <- rep.int(1,nrow(xdat.numeric))
+      res <- rep.int(1,nrow(xdat))
       if(degree[1] > 0) res <- cbind(1, mypoly(xdat.numeric[,1], degree[1]))[, 1 + z[, 1]]
       if(k > 1) for (i in 2:k) if(degree[i] > 0) res <- res * cbind(1, mypoly(xdat.numeric[,i], degree[i]))[, 1 + z[, i]]
       res <- as.matrix(res)
@@ -501,7 +514,7 @@ npregiv <- function(y,
                      degree=NULL,
                      leave.one.out=FALSE,
                      ...) {
-  
+
     ## Don't think this error checking is robust
     
     if(is.null(tydat)) stop("Error: You must provide y data")
@@ -616,6 +629,7 @@ npregiv <- function(y,
       coef.mat <- matrix(maxPenalty,ncol(W),n.eval)
       epsilon <- 1.0/n.eval
       ridge <- double(n.eval)
+      ridge.lc <- double(n.eval)      
       doridge <- !logical(n.eval)
       
       nc <- ncol(tww[,,1])
@@ -625,9 +639,8 @@ npregiv <- function(y,
       
       ridger <- function(i) {
         doridge[i] <<- FALSE
-        ridge.val <- ridge[i]*tyw[1,i][1]/NZD(tww[,,i][1,1])
-        tryCatch(solve(tww[,,i]+diag(rep(ridge[i],nc)),
-                       tyw[,i]+c(ridge.val,rep(0,nc-1))),
+        ridge.lc[i] <- ridge[i]*tyw[1,i][1]/NZD(tww[,,i][1,1])
+        tryCatch(chol2inv(chol(tww[,,i]+diag(rep(ridge[i],nc))))%*%tyw[,i],
                  error = function(e){
                    ridge[i] <<- ridge[i]+epsilon
                    doridge[i] <<- TRUE
@@ -641,7 +654,7 @@ npregiv <- function(y,
       }
       
       mhat <- sapply(1:n.eval, function(i) {
-        W.eval[i,, drop = FALSE] %*% coef.mat[,i]
+        (1-ridge[i])*W.eval[i,, drop = FALSE] %*% coef.mat[,i] + ridge.lc[i]
       })
       
       return(list(mean = mhat,grad = t(coef.mat[-1,])))
@@ -656,9 +669,9 @@ npregiv <- function(y,
                              degree=NULL,
                              W=NULL,
                              ...) {
-  
+    
     ## Don't think this error checking is robust
-  
+    
     if(is.null(ydat)) stop("Error: You must provide y data")
     if(is.null(xdat)) stop("Error: You must provide X data")  
     if(is.null(W)) stop("Error: You must provide a weighting matrix W")
@@ -722,6 +735,7 @@ npregiv <- function(y,
         mean.loo <- rep(maxPenalty,n)
         epsilon <- 1.0/n
         ridge <- double(n)
+        ridge.lc <- double(n)        
         doridge <- !logical(n)
   
         nc <- ncol(tww[,,1])
@@ -731,9 +745,8 @@ npregiv <- function(y,
   
         ridger <- function(i) {
           doridge[i] <<- FALSE
-          ridge.val <- ridge[i]*tyw[1,i][1]/NZD(tww[,,i][1,1])
-          W[i,, drop = FALSE] %*% tryCatch(solve(tww[,,i]+diag(rep(ridge[i],nc)),
-                  tyw[,i]+c(ridge.val,rep(0,nc-1))),
+          ridge.lc[i] <- ridge[i]*tyw[1,i][1]/NZD(tww[,,i][1,1])
+          W[i,, drop = FALSE] %*% tryCatch(chol2inv(chol(tww[,,i]+diag(rep(ridge[i],nc))))%*%tyw[,i],
                   error = function(e){
                     ridge[i] <<- ridge[i]+epsilon
                     doridge[i] <<- TRUE
@@ -743,7 +756,7 @@ npregiv <- function(y,
   
         while(any(doridge)){
           iloo <- (1:n)[doridge]
-          mean.loo[iloo] <- sapply(iloo, ridger)
+          mean.loo[iloo] <- (1-ridge[iloo])*sapply(iloo, ridger) + ridge.lc[iloo]
         }
   
         if (!any(mean.loo == maxPenalty)){
@@ -845,6 +858,7 @@ npregiv <- function(y,
         ghat <- rep(maxPenalty,n)
         epsilon <- 1.0/n
         ridge <- double(n)
+        ridge.lc <- double(n)        
         doridge <- !logical(n)
   
         nc <- ncol(tww[,,1])
@@ -854,9 +868,8 @@ npregiv <- function(y,
   
         ridger <- function(i) {
           doridge[i] <<- FALSE
-          ridge.val <- ridge[i]*tyw[1,i][1]/NZD(tww[,,i][1,1])
-          W[i,, drop = FALSE] %*% tryCatch(solve(tww[,,i]+diag(rep(ridge[i],nc)),
-                  tyw[,i]+c(ridge.val,rep(0,nc-1))),
+          ridge.lc[i] <- ridge[i]*tyw[1,i][1]/NZD(tww[,,i][1,1])
+          W[i,, drop = FALSE] %*% tryCatch(chol2inv(chol(tww[,,i]+diag(rep(ridge[i],nc))))%*%tyw[,i],
                   error = function(e){
                     ridge[i] <<- ridge[i]+epsilon
                     doridge[i] <<- TRUE
@@ -866,15 +879,17 @@ npregiv <- function(y,
   
         while(any(doridge)){
           ii <- (1:n)[doridge]
-          ghat[ii] <- sapply(ii, ridger)
+          ghat[ii] <- (1-ridge[ii])*sapply(ii, ridger) + ridge.lc[ii]
         }
   
         trH <- kernel.i.eq.j*sum(sapply(1:n,function(i){
-          W[i,, drop = FALSE] %*% solve(tww[,,i]+diag(rep(ridge[i],nc))) %*% t(W[i,, drop = FALSE])
+          (1-ridge[i])*W[i,, drop = FALSE] %*% chol2inv(chol(tww[,,i]+diag(rep(ridge[i],nc)))) %*% t(W[i,, drop = FALSE]) + ridge[i]/NZD(tww[,,i][1,1])
         }))
   
-        if (!any(ghat == maxPenalty)){
-          fv <- log(mean((ydat-ghat)^2)) + (1+trH/n)/(1-(trH+2)/n)
+        aic.penalty <- (1+trH/n)/(1-(trH+2)/n)
+  
+        if (!any(ghat == maxPenalty) & (aic.penalty > 0)){
+          fv <- log(mean((ydat-ghat)^2)) + aic.penalty
         } else {
           fv <- maxPenalty
         }
@@ -891,7 +906,7 @@ npregiv <- function(y,
                     xdat=NULL,
                     degree=NULL,
                     bwmethod=c("cv.ls","cv.aic"),
-                    nmulti=NULL,
+                    nmulti=nmulti,
                     random.seed=42,
                     optim.maxattempts = 10,
                     optim.method=c("Nelder-Mead", "BFGS", "CG"),
@@ -943,7 +958,7 @@ npregiv <- function(y,
     ## First initialize initial search values of the vector of
     ## bandwidths to lie in [0,1]
   
-    if(debug) write(c("cv",paste(rep("x",num.bw),seq(1:num.bw),sep="")),file="optim.debug",ncol=(num.bw+1))
+    if(debug) write(c("cv",paste(rep("x",num.bw),seq(1:num.bw),sep="")),file="optim.debug",ncolumns=(num.bw+1))
   
     ## Pass in the local polynomial weight matrix rather than
     ## recomputing with each iteration.
@@ -962,7 +977,7 @@ npregiv <- function(y,
         lscv <- maxPenalty
       }
   
-      if(debug) write(c(lscv,bw.gamma),file="optim.debug",ncol=(num.bw+1),append=TRUE)
+      if(debug) write(c(lscv,bw.gamma),file="optim.debug",ncolumns=(num.bw+1),append=TRUE)
       return(lscv)
     }
   
@@ -978,7 +993,7 @@ npregiv <- function(y,
         aicc <- maxPenalty
       }
   
-      if(debug) write(c(aicc,bw.gamma),file="optim.debug",ncol=(num.bw+1),append=TRUE)
+      if(debug) write(c(aicc,bw.gamma),file="optim.debug",ncolumns=(num.bw+1),append=TRUE)
       return(aicc)
     }
   
@@ -998,7 +1013,7 @@ npregiv <- function(y,
   
       for(i in 1:ncol(xdat)) {
         if(xdat.numeric[i]==TRUE) {
-          init.search.vals[i] <- runif(1,.5,1.5)*(IQR(xdat[,i])/1.349)*nrow(xdat)^{-1/(4+num.numeric)}
+          init.search.vals[i] <- runif(1,.5,1.5)*EssDee(xdat[,i])*nrow(xdat)^{-1/(4+num.numeric)}
         }
       }
   
@@ -1025,7 +1040,7 @@ npregiv <- function(y,
         while((optim.return$convergence != 0) && (attempts <= optim.maxattempts)) {
           init.search.vals <- runif(ncol(xdat),0,1)
           if(xdat.numeric[i]==TRUE) {
-            init.search.vals[i] <- runif(1,.5,1.5)*(IQR(xdat[,i])/1.349)*nrow(xdat)^{-1/(4+num.numeric)}
+            init.search.vals[i] <- runif(1,.5,1.5)*EssDee(xdat[,i])*nrow(xdat)^{-1/(4+num.numeric)}
           }
           attempts <- attempts + 1
           optim.control$abstol <- optim.control$abstol * 10.0
@@ -1054,7 +1069,7 @@ npregiv <- function(y,
         while((optim.return$convergence != 0) && (attempts <= optim.maxattempts)) {
           init.search.vals <- runif(ncol(xdat),0,1)
           if(xdat.numeric[i]==TRUE) {
-            init.search.vals[i] <- runif(1,.5,1.5)*(IQR(xdat[,i])/1.349)*nrow(xdat)^{-1/(4+num.numeric)}
+            init.search.vals[i] <- runif(1,.5,1.5)*EssDee(xdat[,i])*nrow(xdat)^{-1/(4+num.numeric)}
           }
           attempts <- attempts + 1
           optim.control$abstol <- optim.control$abstol * 10.0
@@ -1080,11 +1095,11 @@ npregiv <- function(y,
         best <- iMulti
         if(debug) {
           if(iMulti==1) {
-            write(cbind(iMulti,t(bw.opt)),"bandwidth.out",ncol=(1+length(bw.opt)))
-            write(cbind(iMulti,fv),"optim.out",ncol=2)
+            write(cbind(iMulti,t(bw.opt)),"bandwidth.out",ncolumns=(1+length(bw.opt)))
+            write(cbind(iMulti,fv),"optim.out",ncolumns=2)
           } else {
-            write(cbind(iMulti,t(bw.opt)),"bandwidth.out",ncol=(1+length(bw.opt)),append=TRUE)
-            write(cbind(iMulti,fv),"optim.out",ncol=2,append=TRUE)
+            write(cbind(iMulti,t(bw.opt)),"bandwidth.out",ncolumns=(1+length(bw.opt)),append=TRUE)
+            write(cbind(iMulti,fv),"optim.out",ncolumns=2,append=TRUE)
           }
         }
       }
@@ -1104,6 +1119,10 @@ npregiv <- function(y,
   console <- newLineConsole()
     
   ## Basic error checking
+
+  if(!is.logical(penalize.iteration)) stop("penalize.iteration must be logical (TRUE/FALSE)")    
+  if(!is.logical(smooth.residuals)) stop("smooth.residuals must be logical (TRUE/FALSE)")  
+  if(!is.logical(stop.on.increase)) stop("stop.on.increase must be logical (TRUE/FALSE)")  
   
   if(missing(y)) stop("You must provide y")
   if(missing(z)) stop("You must provide z")
@@ -1111,17 +1130,20 @@ npregiv <- function(y,
   if(NCOL(y) > 1) stop("y must be univariate")
   if(NROW(y) != NROW(z) || NROW(y) != NROW(w)) stop("y, z, and w have differing numbers of rows")
   if(iterate.max < 2) stop("iterate.max must be at least 2")
+  if(iterate.diff.tol < 0) stop("iterate.diff.tol must be non-negative")
+  if(constant <= 0 || constant >=1) stop("constant must lie in (0,1)")
   if(p < 0) stop("p must be a non-negative integer")
 
   if(!is.null(alpha) && alpha <= 0) stop("alpha must be positive")
 
+  start.from <- match.arg(start.from)
   method <- match.arg(method)
   
   ## Check for evaluation data
   
   if(is.null(zeval)) zeval <- z
   if(is.null(weval)) weval <- w
-  if(is.null(weval)) xeval <- x
+  if(is.null(xeval)) xeval <- x
 
   ## Need to determine how many x, w, z are numeric
 
@@ -1142,28 +1164,80 @@ npregiv <- function(y,
     ## bandwidths, one for y on w and one for phi(z) on w (in the
     ## first step we use E(y|w) as a proxy for phi(z) and use
     ## bandwidths for y on w).
+
+    ## Convergence value returned for Landweber-Fridman but value
+    ## required
+
+    convergence <- NULL
     
     console <- printClear(console)
     console <- printPop(console)
     console <- printPush("Computing bandwidths for E(y|w)...", console)
-    hyw <- glpcv(ydat=y, xdat=w, degree=rep(p, num.w.numeric),...)
+
+    hyw <- glpcv(ydat=y,
+                 xdat=w,
+                 degree=rep(p, num.w.numeric),
+                 nmulti=nmulti,
+                 random.seed=random.seed,
+                 optim.maxattempts=optim.maxattempts,
+                 optim.method=optim.method,
+                 optim.reltol=optim.reltol,
+                 optim.abstol=optim.abstol,
+                 optim.maxit=optim.maxit,
+                 ...)
+
     console <- printClear(console)
     console <- printPop(console)
     console <- printPush("Computing weight matrix and E(y|w)...", console)  
-    E.y.w <- glpreg(tydat=y, txdat=w, exdat=weval, bws=hyw$bw, degree=rep(p, num.w.numeric),...)$mean
-    KYW <- Kmat.lp(mydata.train=data.frame(w), mydata.eval=data.frame(w=weval), bws=hyw$bw, p=rep(p, num.w.numeric))
+
+    E.y.w <- glpreg(tydat=y,
+                    txdat=w,
+                    exdat=weval,
+                    bws=hyw$bw,
+                    degree=rep(p, num.w.numeric),
+                    ...)$mean
+
+    KYW <- Kmat.lp(mydata.train=data.frame(w),
+                   mydata.eval=data.frame(w=weval),
+                   bws=hyw$bw,
+                   p=rep(p, num.w.numeric),
+                   ...)
     
     ## We conduct local polynomial kernel regression of E(y|w) on z
     
     console <- printClear(console)
     console <- printPop(console)
     console <- printPush("Computing bandwidths for E(E(y|w)|z)...", console)
-    hywz <- glpcv(ydat=E.y.w, xdat=z, degree=rep(p, num.z.numeric),...)
+
+    hywz <- glpcv(ydat=E.y.w,
+                  xdat=z,
+                  degree=rep(p, num.z.numeric),
+                  nmulti=nmulti,
+                  random.seed=random.seed,
+                  optim.maxattempts=optim.maxattempts,
+                  optim.method=optim.method,
+                  optim.reltol=optim.reltol,
+                  optim.abstol=optim.abstol,
+                  optim.maxit=optim.maxit,
+                  ...)
+
     console <- printClear(console)
     console <- printPop(console)
     console <- printPush("Computing weight matrix and E(E(y|w)|z)...", console)  
-    E.E.y.w.z <- glpreg(tydat=E.y.w, txdat=z, eydat=E.y.w, exdat=zeval, bws=hywz$bw, degree=rep(p, num.z.numeric),...)$mean
-    KYWZ <- Kmat.lp(mydata.train=data.frame(z), mydata.eval=data.frame(z=zeval), bws=hywz$bw, p=rep(p, num.z.numeric))
+
+    E.E.y.w.z <- glpreg(tydat=E.y.w,
+                        txdat=z,
+                        eydat=E.y.w,
+                        exdat=zeval,
+                        bws=hywz$bw,
+                        degree=rep(p, num.z.numeric),
+                        ...)$mean
+
+    KYWZ <- Kmat.lp(mydata.train=data.frame(z),
+                    mydata.eval=data.frame(z=zeval),
+                    bws=hywz$bw,
+                    p=rep(p, num.z.numeric),
+                    ...)
     
     ## Next, we minimize the function ittik to obtain the optimal value
     ## of alpha (here we use the iterated Tikhonov function) to
@@ -1187,7 +1261,7 @@ npregiv <- function(y,
     console <- printClear(console)
     console <- printPop(console)
     console <- printPush("Computing initial phi(z) estimate...", console)
-    phihat <- as.vector(tikh(alpha, CZ = KYW, CY = KYWZ, Cr.r = E.E.y.w.z))
+    phi <- as.vector(tikh(alpha, CZ = KYW, CY = KYWZ, Cr.r = E.E.y.w.z))
     
     ## KYWZ and KYWS no longer used, save memory
     
@@ -1196,21 +1270,62 @@ npregiv <- function(y,
     console <- printClear(console)
     console <- printPop(console)
     console <- printPush("Computing bandwidths for E(phi(z)|w)...", console)
-    hphiw <- glpcv(ydat=phihat, xdat=w, degree=rep(p, num.w.numeric),...)
+
+    hphiw <- glpcv(ydat=phi,
+                   xdat=w,
+                   degree=rep(p, num.w.numeric),
+                   nmulti=nmulti,
+                   random.seed=random.seed,
+                   optim.maxattempts=optim.maxattempts,
+                   optim.method=optim.method,
+                   optim.reltol=optim.reltol,
+                   optim.abstol=optim.abstol,
+                   optim.maxit=optim.maxit,
+                   ...)
+
     console <- printClear(console)
     console <- printPop(console)
     console <- printPush("Computing weight matrix for E(phi(z)|w)...", console)
-    E.phihat.w <- glpreg(tydat=phihat, txdat=w, eydat=phihat, exdat=weval, bws=hphiw$bw, degree=rep(p, num.w.numeric),...)$mean
-    KPHIW <- Kmat.lp(mydata.train=data.frame(w), mydata.eval=data.frame(w=weval), bws=hphiw$bw, p=rep(p, num.w.numeric))
+
+    E.phi.w <- glpreg(tydat=phi,
+                         txdat=w,
+                         eydat=phi,
+                         exdat=weval,
+                         bws=hphiw$bw,
+                         degree=rep(p, num.w.numeric),
+                         ...)$mean
+
+    KPHIW <- Kmat.lp(mydata.train=data.frame(w),
+                     mydata.eval=data.frame(w=weval),
+                     bws=hphiw$bw,
+                     p=rep(p, num.w.numeric),
+                     ...)
     
     console <- printClear(console)
     console <- printPop(console)
     console <- printPush("Computing bandwidths for E(E(phi(z)|w)|z)...", console)
-    hphiwz <- glpcv(ydat=E.phihat.w, xdat=z, degree=rep(p, num.z.numeric))
+
+    hphiwz <- glpcv(ydat=E.phi.w,
+                    xdat=z,
+                    degree=rep(p, num.z.numeric),
+                    nmulti=nmulti,
+                    random.seed=random.seed,
+                    optim.maxattempts=optim.maxattempts,
+                    optim.method=optim.method,
+                    optim.reltol=optim.reltol,
+                    optim.abstol=optim.abstol,
+                    optim.maxit=optim.maxit,
+                    ...)
+
     console <- printClear(console)
     console <- printPop(console)
     console <- printPush("Computing weight matrix for E(E(phi(z)|w)|z)...", console)
-    KPHIWZ <- Kmat.lp(mydata.train=data.frame(z), mydata.eval=data.frame(z=zeval), bws=hphiwz$bw, p=rep(p, num.z.numeric))
+
+    KPHIWZ <- Kmat.lp(mydata.train=data.frame(z),
+                      mydata.eval=data.frame(z=zeval),
+                      bws=hphiwz$bw,
+                      p=rep(p, num.z.numeric),
+                      ...)
     
     ## Next, we minimize the function ittik to obtain the optimal value
     ## of alpha (here we use the iterated Tikhonov approach) to
@@ -1229,7 +1344,7 @@ npregiv <- function(y,
     console <- printClear(console)
     console <- printPop(console)
     console <- printPush("Computing final phi(z) estimate...", console)
-    phihat <- as.vector(tikh(alpha, CZ = KPHIW, CY = KPHIWZ, Cr.r = E.E.y.w.z))
+    phi <- as.vector(tikh(alpha, CZ = KPHIW, CY = KPHIWZ, Cr.r = E.E.y.w.z))
     
     console <- printClear(console)
     console <- printPop(console)
@@ -1237,12 +1352,39 @@ npregiv <- function(y,
     if((alpha-alpha.min)/alpha.min < 0.01) warning(paste("Tikhonov parameter alpha (",formatC(alpha,digits=4,format="f"),") is close to the search minimum (",alpha.min,")",sep=""))
     if((alpha.max-alpha)/alpha.max < 0.01) warning(paste("Tikhonov parameter alpha (",formatC(alpha,digits=4,format="f"),") is close to the search maximum (",alpha.max,")",sep=""))
     
-    return(list(phihat=phihat, alpha=alpha))
+    return(list(phi=phi, alpha=alpha))
     
   } else {
 
     ## Landweber-Fridman
 
+    ## For the stopping rule
+    
+    console <- printClear(console)
+    console <- printPop(console)
+    console <- printPush(paste("Computing bandwidths and E(y|w) for stopping rule...",sep=""),console)
+
+    norm.stop <- numeric()
+
+    h <- glpcv(ydat=y,
+               xdat=w,
+               degree=rep(p, num.w.numeric),
+               nmulti=nmulti,
+               random.seed=random.seed,
+               optim.maxattempts=optim.maxattempts,
+               optim.method=optim.method,
+               optim.reltol=optim.reltol,
+               optim.abstol=optim.abstol,
+               optim.maxit=optim.maxit,
+               ...)
+
+    E.y.w <- glpreg(tydat=y,
+                    txdat=w,
+                    exdat=weval,
+                    bws=h$bw,
+                    degree=rep(p, num.w.numeric),
+                    ...)$mean
+
     ## We begin the iteration computing phi.0 and phi.1 directly, then
     ## interate.
     
@@ -1250,42 +1392,116 @@ npregiv <- function(y,
     console <- printPop(console)
     console <- printPush(paste("Computing bandwidths and E(y|z) for iteration 0...",sep=""),console)
 
-    h <- glpcv(ydat=y, xdat=z, degree=rep(p, num.z.numeric),...)
-    phi.0 <- glpreg(tydat=y, txdat=z, exdat=zeval, bws=h$bw, degree=rep(p, num.z.numeric),...)$mean
+    if(is.null(starting.values)) { 
+
+      h <- glpcv(ydat=if(start.from=="Eyz") y else E.y.w,
+                 xdat=z,
+                 degree=rep(p, num.z.numeric),
+                 nmulti=nmulti,
+                 random.seed=random.seed,
+                 optim.maxattempts=optim.maxattempts,
+                 optim.method=optim.method,
+                 optim.reltol=optim.reltol,
+                 optim.abstol=optim.abstol,
+                 optim.maxit=optim.maxit,
+                 ...)
+
+      phi.0 <- glpreg(tydat=if(start.from=="Eyz") y else E.y.w,
+                      txdat=z,
+                      exdat=zeval,
+                      bws=h$bw,
+                      degree=rep(p, num.z.numeric),
+                      ...)$mean
+
+    } else {
+
+      ## Starting values input by user
+      
+      phi.0 <- starting.values
+
+    }
+
+    starting.values.phi <- phi.0
     
     console <- printClear(console)
     console <- printPop(console)
     console <- printPush(paste("Computing bandwidths and E(y-phi(z)|w) for iteration 1...",sep=""),console)
 
-    resid <- y - phi.0
-    h <- glpcv(ydat=resid, xdat=w, degree=rep(p, num.w.numeric),...)
-    resid.fitted <- glpreg(tydat=resid, txdat=w, eydat=resid, exdat=weval, bws=h$bw, degree=rep(p, num.w.numeric),...)$mean
+    if(smooth.residuals) {
+
+      resid <- y - phi.0
+      
+      h <- glpcv(ydat=resid,
+                 xdat=w,
+                 degree=rep(p, num.w.numeric),
+                 nmulti=nmulti,
+                 random.seed=random.seed,
+                 optim.maxattempts=optim.maxattempts,
+                 optim.method=optim.method,
+                 optim.reltol=optim.reltol,
+                 optim.abstol=optim.abstol,
+                 optim.maxit=optim.maxit,
+                 ...)
+      
+      resid.fitted <- glpreg(tydat=resid,
+                             txdat=w,
+                             eydat=resid,
+                             exdat=weval,
+                             bws=h$bw,
+                             degree=rep(p, num.w.numeric),
+                             ...)$mean
+      
+    } else {
+      
+      h <- glpcv(ydat=phi.0,
+                 xdat=w,
+                 degree=rep(p, num.w.numeric),
+                 nmulti=nmulti,
+                 random.seed=random.seed,
+                 optim.maxattempts=optim.maxattempts,
+                 optim.method=optim.method,
+                 optim.reltol=optim.reltol,
+                 optim.abstol=optim.abstol,
+                 optim.maxit=optim.maxit,
+                 ...)
+      
+      resid.fitted <- E.y.w - glpreg(tydat=phi.0,
+                                    txdat=w,
+                                    eydat=resid,
+                                    exdat=weval,
+                                    bws=h$bw,
+                                    degree=rep(p, num.w.numeric),
+                                    ...)$mean
+      
+    }
+
+    norm.stop[1] <- sum(resid.fitted^2)/sum(E.y.w^2)
 
     console <- printClear(console)
     console <- printPop(console)
     console <- printPush(paste("Computing bandwidths and E(E(y-phi(z)|w)|z) for iteration 1...",sep=""),console)
 
-    h <- glpcv(ydat=resid.fitted, xdat=z, degree=rep(p, num.z.numeric),...)
-    phi.j.m.1 <- phi.0 + glpreg(tydat=resid.fitted, txdat=z, eydat=resid.fitted, exdat=zeval, bws=h$bw, degree=rep(p, num.z.numeric),...)$mean
+    h <- glpcv(ydat=resid.fitted,
+               xdat=z,
+               degree=rep(p, num.z.numeric),
+               nmulti=nmulti,
+               random.seed=random.seed,
+               optim.maxattempts=optim.maxattempts,
+               optim.method=optim.method,
+               optim.reltol=optim.reltol,
+               optim.abstol=optim.abstol,
+               optim.maxit=optim.maxit,
+               ...)
+
+    phi <- phi.0 + constant*glpreg(tydat=resid.fitted,
+                                         txdat=z,
+                                         eydat=resid.fitted,
+                                         exdat=zeval,
+                                         bws=h$bw,
+                                         degree=rep(p, num.z.numeric),
+                                         ...)$mean
     
-    ## For the stopping rule
-    
-    console <- printClear(console)
-    console <- printPop(console)
-    console <- printPush(paste("Computing bandwidths and E(y|w) for stopping rule...",sep=""),console)
-
-    norm.stop <- numeric()
-    h <- glpcv(ydat=y, xdat=w, degree=rep(p, num.w.numeric),...)
-    E.y.w <- glpreg(tydat=y, txdat=w, exdat=weval, bws=h$bw, degree=rep(p, num.w.numeric),...)$mean
-    phihat <- phi.j.m.1
-
-    console <- printClear(console)
-    console <- printPop(console)
-    console <- printPush(paste("Computing bandwidths and E(phi(z)|w) for stopping rule...",sep=""),console)
-
-    h.E.phi.w <- glpcv(ydat=phi.j.m.1, xdat=w, degree=rep(p, num.w.numeric),...)
-    E.phi.w <- glpreg(tydat=phi.j.m.1, txdat=w, eydat=phi.j.m.1, exdat=weval, bws=h.E.phi.w$bw, degree=rep(p, num.w.numeric),...)$mean
-    norm.stop[1] <- mean(((E.y.w-E.phi.w)/E.y.w)^2)
+    phi.mat <- phi
 
     for(j in 2:iterate.max) {
 
@@ -1293,38 +1509,154 @@ npregiv <- function(y,
       console <- printPop(console)
       console <- printPush(paste("Computing bandwidths and E(y-phi(z)|w) for iteration ", j,"...",sep=""),console)
 
-      resid <- y - phi.j.m.1
-      h.resid.w <- glpcv(ydat=resid, xdat=w, degree=rep(p, num.w.numeric),...)
-      resid.fitted <- glpreg(tydat=resid, txdat=w, eydat=resid, exdat=weval, bws=h.resid.w$bw, degree=rep(p, num.w.numeric),...)$mean
+      if(smooth.residuals) {
+
+        resid <- y - phi
+        h.resid.w <- glpcv(ydat=resid,
+                           xdat=w,
+                           degree=rep(p, num.w.numeric),
+                           nmulti=nmulti,
+                           random.seed=random.seed,
+                           optim.maxattempts=optim.maxattempts,
+                           optim.method=optim.method,
+                           optim.reltol=optim.reltol,
+                           optim.abstol=optim.abstol,
+                           optim.maxit=optim.maxit,
+                           ...)
+        
+        resid.fitted <- glpreg(tydat=resid,
+                               txdat=w,
+                               eydat=resid,
+                               exdat=weval,
+                               bws=h.resid.w$bw,
+                               degree=rep(p, num.w.numeric),
+                               ...)$mean
+
+      } else {
+
+        h <- glpcv(ydat=phi,
+                   xdat=w,
+                   degree=rep(p, num.w.numeric),
+                   nmulti=nmulti,
+                   random.seed=random.seed,
+                   optim.maxattempts=optim.maxattempts,
+                   optim.method=optim.method,
+                   optim.reltol=optim.reltol,
+                   optim.abstol=optim.abstol,
+                   optim.maxit=optim.maxit,
+                   ...)
+      
+        resid.fitted <- E.y.w - glpreg(tydat=phi,
+                                       txdat=w,
+                                       eydat=resid,
+                                       exdat=weval,
+                                       bws=h$bw,
+                                       degree=rep(p, num.w.numeric),
+                                       ...)$mean
+        
+      }
+
+      norm.stop[j] <- ifelse(penalize.iteration,j*sum(resid.fitted^2)/sum(E.y.w^2),sum(resid.fitted^2)/sum(E.y.w^2))
 
       console <- printClear(console)
       console <- printPop(console)
       console <- printPush(paste("Computing bandwidths and E(E(y-phi(z)|w)|z) for iteration ", j,"...",sep=""),console)
 
-      h.resid.fitted.z <- glpcv(ydat=resid.fitted, xdat=z, degree=rep(p, num.z.numeric),...)
-      phi.j <- phi.j.m.1 + constant*glpreg(tydat=resid.fitted, txdat=z, eydat=resid.fitted, exdat=zeval, bws=h.resid.fitted.z$bw, degree=rep(p, num.z.numeric),...)$mean
-      phi.j.m.1 <- phi.j
+      h.resid.fitted.z <- glpcv(ydat=resid.fitted,
+                                xdat=z,
+                                degree=rep(p, num.z.numeric),
+                                nmulti=nmulti,
+                                random.seed=random.seed,
+                                optim.maxattempts=optim.maxattempts,
+                                optim.method=optim.method,
+                                optim.reltol=optim.reltol,
+                                optim.abstol=optim.abstol,
+                                optim.maxit=optim.maxit,
+                                ...)
+
+      phi <- phi + constant*glpreg(tydat=resid.fitted,
+                                           txdat=z,
+                                           eydat=resid.fitted,
+                                           exdat=zeval,
+                                           bws=h.resid.fitted.z$bw,
+                                           degree=rep(p, num.z.numeric),
+                                           ...)$mean
+
+      phi.mat <- cbind(phi.mat,phi)
 
       console <- printClear(console)
       console <- printPop(console)
       console <- printPush(paste("Computing stopping rule for iteration ", j,"...",sep=""),console)
 
-      ## For the stopping rule (use same smoothing as original)
-      E.phi.w <- glpreg(tydat=phi.j, txdat=w, eydat=phi.j, exdat=weval, bws=h.E.phi.w$bw, degree=rep(p, num.w.numeric),...)$mean
-      norm.stop[j] <- mean(((E.y.w-E.phi.w)/E.y.w)^2)
+      ## The number of iterations in LF is asymptotically equivalent
+      ## to 1/alpha (where alpha is the regularization parameter in
+      ## Tikhonov).  Plus the criterion function we use is increasing
+      ## for very small number of iterations. So we need a threshold
+      ## after which we can pretty much confidently say that the
+      ## stopping criterion is decreasing.  In Darolles et al. (2011)
+      ## \alpha ~ O(N^(-1/(min(beta,2)+2)), where beta is the so
+      ## called qualification of your regularization method. Take the
+      ## worst case in which beta = 0 and then the number of
+      ## iterations is ~ N^0.5.
 
-      ## If objective increases or we are below stopping tolerance then break
+      if(j > round(sqrt(nrow(z))) && !is.monotone.increasing(norm.stop)) {
 
-      if((norm.stop[j] > norm.stop[j-1]) || ((norm.stop[j-1]-norm.stop[j]) < iterate.tol)) break()
+        ## If stopping rule criterion increases or we are below stopping
+        ## tolerance then break
+        
+        if(stop.on.increase && norm.stop[j] > norm.stop[j-1]) {
+          convergence <- "STOP_ON_INCREASE"
+          break()
+        }
+        if(abs(norm.stop[j-1]-norm.stop[j]) < iterate.diff.tol) {
+          convergence <- "ITERATE_DIFF_TOL"
+          break()
+        }
+        
+      }
+
+      convergence <- "ITERATE_MAX"
 
     }
 
+    ## Extract minimum, and check for monotone increasing function and
+    ## issue warning in that case. Otherwise allow for an increasing
+    ## then decreasing (and potentially increasing thereafter) portion
+    ## of the stopping function, ignore the initial increasing portion,
+    ## and take the min from where the initial inflection point occurs
+    ## to the length of norm.stop
+
+    norm.value <- norm.stop/(1:length(norm.stop))
+
+    if(which.min(norm.stop) == 1 && is.monotone.increasing(norm.stop)) {
+      warning("Stopping rule increases monotonically (consult model$norm.stop):\nThis could be the result of an inspired initial value (unlikely)\nNote: we suggest manually choosing phi.0 and restarting (e.g. instead set `starting.values' to E[E(Y|w)|z])")
+      convergence <- "FAILURE_MONOTONE_INCREASING"
+#      phi <- starting.values.phi
+      j <- 1
+      while(norm.value[j+1] > norm.value[j]) j <- j + 1
+      j <- j-1 + which.min(norm.value[j:length(norm.value)])
+      phi <- phi.mat[,j]
+    } else {
+      ## Ignore the initial increasing portion, take the min to the
+      ## right of where the initial inflection point occurs
+      j <- 1
+      while(norm.stop[j+1] > norm.stop[j]) j <- j + 1
+      j <- j-1 + which.min(norm.stop[j:length(norm.stop)])
+      phi <- phi.mat[,j]
+    }    
+
     console <- printClear(console)
     console <- printPop(console)
 
     if(j == iterate.max) warning("iterate.max reached: increase iterate.max or inspect norm.stop vector")
 
-    return(list(phihat=phi.j, num.iterations=j, norm.stop=norm.stop))
+    return(list(phi=phi,
+                phi.mat=phi.mat,
+                num.iterations=j,
+                norm.stop=norm.stop,
+                norm.value=norm.value,
+                convergence=convergence,
+                starting.values.phi=starting.values.phi))
 
   }
   
